'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var animationFrameChunkScheduler = typeof window !== 'undefined' &&
    typeof window.requestAnimationFrame === 'function' &&
    typeof window.cancelAnimationFrame === 'function' ?
    {
        request: function (fn) { return window.requestAnimationFrame(fn); },
        cancel: function (token) { return window.cancelAnimationFrame(token); }
    } :
    null;

var idleCallbackChunkScheduler = typeof window !== 'undefined' &&
    typeof window.requestIdleCallback === 'function' &&
    typeof window.cancelIdleCallback === 'function' ?
    {
        request: function (fn) { return window.requestIdleCallback(fn); },
        cancel: function (token) { return window.cancelIdleCallback(token); }
    } :
    null;

var immediateChunkScheduler = typeof setImmediate === 'function' &&
    typeof clearImmediate === 'function' ?
    {
        request: function (fn) { return setImmediate(fn); },
        cancel: function (token) { return clearImmediate(token); }
    } :
    null;

var timeoutChunkScheduler = {
    request: function (fn) { return setTimeout(fn, 0); },
    cancel: function (token) { return clearTimeout(token); }
};

var BUILTIN_CHUNK_SHEDULERS = {
    auto: idleCallbackChunkScheduler ||
        animationFrameChunkScheduler ||
        immediateChunkScheduler,
    animationFrame: animationFrameChunkScheduler,
    idleCallback: idleCallbackChunkScheduler,
    immediate: immediateChunkScheduler,
    timeout: timeoutChunkScheduler
};
function getChunkScheduler(type) {
    return typeof type === 'string' ?
        BUILTIN_CHUNK_SHEDULERS[type] || timeoutChunkScheduler :
        type;
}

var now = typeof performance === 'object' && typeof performance.now === 'function' ?
    function () { return performance.now(); } :
    (function () {
        try {
            return require('perf_hooks').performance.now;
        }
        catch (_a) {
            return Date.now;
        }
    })();
var microtaskPromise = Promise.resolve();
function microtask(fn) {
    microtaskPromise.then(fn);
}

function createScheduler(_a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.chunkScheduler, chunkSchedulerType = _c === void 0 ? 'auto' : _c, _d = _b.chunkBudget, chunkBudget = _d === void 0 ? 10 : _d;
    var pendingTasks = new Map();
    var chunkScheduler = getChunkScheduler(chunkSchedulerType);
    var chunkSchedulerToken = null;
    var tasksOrder = [];
    function chunk() {
        chunkSchedulerToken = null;
        var checkBudget = false;
        var startTime = now();
        var restBudget = chunkBudget;
        var nextTasksOrder = [];
        while (tasksOrder.length > 0) {
            var promise = tasksOrder.shift();
            var task = pendingTasks.get(promise);
            var unitExecuted = false;
            if (checkBudget && restBudget < task.meanUnitElapsedTime) {
                nextTasksOrder.push(promise);
            }
            else {
                checkBudget = true;
                try {
                    var _a = task.nextUnit(task.result), next = _a.next, result = _a.result;
                    unitExecuted = true;
                    task.nextUnit = next;
                    task.result = result;
                    if (next === null) {
                        pendingTasks.delete(promise);
                        task.resolve(result);
                    }
                    else {
                        tasksOrder.push(promise);
                    }
                }
                catch (e) {
                    pendingTasks.delete(promise);
                    task.reject(e);
                }
            }
            var elapsedTime = now() - startTime;
            if (unitExecuted) {
                task.executedUnitCount++;
                task.totalElapsedTime += elapsedTime;
                task.meanUnitElapsedTime = task.totalElapsedTime / task.executedUnitCount;
            }
            restBudget -= elapsedTime;
            startTime += elapsedTime;
        }
        if (nextTasksOrder.length > 0) {
            tasksOrder = nextTasksOrder;
            chunkSchedulerToken = chunkScheduler.request(chunk);
        }
    }
    return {
        runTask: function (unit) {
            var task;
            var promise = new Promise(function (resolve, reject) {
                task = {
                    nextUnit: unit,
                    result: undefined,
                    executedUnitCount: 0,
                    totalElapsedTime: 0,
                    meanUnitElapsedTime: 0,
                    resolve: resolve,
                    reject: reject
                };
            });
            pendingTasks.set(promise, task);
            microtask(function () {
                // check if it's not already aborted
                if (!pendingTasks.has(promise)) {
                    return;
                }
                tasksOrder.push(promise);
                if (tasksOrder.length === 1) {
                    chunkSchedulerToken = chunkScheduler.request(chunk);
                }
            });
            return promise;
        },
        abortTask: function (promise) {
            if (pendingTasks.delete(promise)) {
                var taskOrderIdx = tasksOrder.indexOf(promise);
                // task can be absent if it's added to pending tasks via `runTask` but then
                // `abortTask` is called synchronously before microtask callback is invoked
                if (taskOrderIdx > -1) {
                    tasksOrder.splice(taskOrderIdx, 1);
                    if (tasksOrder.length === 0 && chunkScheduler.cancel && chunkSchedulerToken !== null) {
                        chunkScheduler.cancel(chunkSchedulerToken);
                        chunkSchedulerToken = null;
                    }
                }
            }
        }
    };
}

exports.createScheduler = createScheduler;
